@@ -12,17 +12,13 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 # ==============================================================================
-
 """Tensor utility functions."""
-from __future__ import absolute_import
-from __future__ import division
-from __future__ import print_function
-
 import collections
 import functools
 import inspect
 import re
 
+from tensorflow.python.framework import strict_mode
 from tensorflow.python.platform import tf_logging as logging
 from tensorflow.python.util import decorator_utils
 from tensorflow.python.util import is_in_graph_mode
@@ -31,7 +27,6 @@ from tensorflow.python.util import tf_decorator
 from tensorflow.python.util import tf_inspect
 from tensorflow.tools.docs import doc_controls
 
-
 # Allow deprecation warnings to be silenced temporarily with a context manager.
 _PRINT_DEPRECATION_WARNINGS = True
 
@@ -39,21 +34,37 @@ _PRINT_DEPRECATION_WARNINGS = True
 _PRINTED_WARNING = {}
 
 
-class DeprecatedNamesAlreadySet(Exception):
+class DeprecatedNamesAlreadySetError(Exception):
   """Raised when setting deprecated names multiple times for the same symbol."""
-  pass
+
+
+def _log_deprecation(msg, *args, **kwargs):
+  """Raises errors for deprecated methods if in strict mode, warns otherwise."""
+  if strict_mode.STRICT_MODE:
+    logging.error(msg, *args, **kwargs)
+    raise RuntimeError(
+        'This behavior has been deprecated, which raises an error in strict'
+        ' mode.'
+    )
+  else:
+    logging.warning(msg, *args, **kwargs)
 
 
 def _add_deprecated_function_notice_to_docstring(doc, date, instructions):
   """Adds a deprecation notice to a docstring for deprecated functions."""
-  main_text = ['THIS FUNCTION IS DEPRECATED. It will be removed %s.' %
-               ('in a future version' if date is None else ('after %s' % date))]
+  main_text = [
+      'THIS FUNCTION IS DEPRECATED. It will be removed %s.'
+      % ('in a future version' if date is None else ('after %s' % date))
+  ]
   if instructions:
     main_text.append('Instructions for updating:')
   return decorator_utils.add_notice_to_docstring(
-      doc, instructions,
+      doc,
+      instructions,
       'DEPRECATED FUNCTION',
-      '(deprecated)', main_text)
+      '(deprecated)',
+      main_text,
+      notice_type='Deprecated')
 
 
 def _add_deprecated_arg_notice_to_docstring(doc, date, instructions,
@@ -63,13 +74,16 @@ def _add_deprecated_arg_notice_to_docstring(doc, date, instructions,
   deprecation_string = ', '.join(sorted(deprecated_names))
 
   return decorator_utils.add_notice_to_docstring(
-      doc, instructions, 'DEPRECATED FUNCTION ARGUMENTS',
+      doc,
+      instructions,
+      'DEPRECATED FUNCTION ARGUMENTS',
       '(deprecated arguments)', [
           'SOME ARGUMENTS ARE DEPRECATED: `(%s)`. '
           'They will be removed %s.' %
           (deprecation_string, 'in a future version' if date is None else
            ('after %s' % date)), 'Instructions for updating:'
-      ])
+      ],
+      notice_type='Deprecated')
 
 
 def _add_deprecated_arg_value_notice_to_docstring(doc, date, instructions,
@@ -83,12 +97,15 @@ def _add_deprecated_arg_value_notice_to_docstring(doc, date, instructions,
   when = 'in a future version' if date is None else ('after %s' % date)
 
   return decorator_utils.add_notice_to_docstring(
-      doc, instructions, 'DEPRECATED FUNCTION ARGUMENT VALUES',
+      doc,
+      instructions,
+      'DEPRECATED FUNCTION ARGUMENT VALUES',
       '(deprecated argument values)', [
           'SOME ARGUMENT VALUES ARE DEPRECATED: `(%s)`. '
-          'They will be removed %s.' % (deprecation_string, when),
-          'Instructions for updating:'
-      ])
+          'They will be removed %s.' %
+          (deprecation_string, when), 'Instructions for updating:'
+      ],
+      notice_type='Deprecated')
 
 
 def _validate_deprecation_args(date, instructions):
@@ -119,7 +136,7 @@ def _safe_eq(a, b):
   return a == b
 
 
-def _wrap_decorator(wrapped_function):
+def _wrap_decorator(wrapped_function, decorator_name):
   """Indicate that one function wraps another.
 
   This decorator wraps a function using `tf_decorator.make_decorator`
@@ -130,13 +147,17 @@ def _wrap_decorator(wrapped_function):
 
   Args:
     wrapped_function: The function that decorated function wraps.
+    decorator_name: The name of the decorator.
 
   Returns:
     Function that accepts wrapper function as an argument and returns
     `TFDecorator` instance.
   """
+
   def wrapper(wrapper_func):
-    return tf_decorator.make_decorator(wrapped_function, wrapper_func)
+    return tf_decorator.make_decorator(wrapped_function, wrapper_func,
+                                       decorator_name)
+
   return wrapper
 
 
@@ -199,14 +220,16 @@ def deprecated_alias(deprecated_name, name, func_or_class, warn_once=True):
     # Make a new class with __init__ wrapped in a warning.
     class _NewClass(func_or_class):  # pylint: disable=missing-docstring
       __doc__ = decorator_utils.add_notice_to_docstring(
-          func_or_class.__doc__, 'Please use %s instead.' % name,
+          func_or_class.__doc__,
+          'Please use %s instead.' % name,
           'DEPRECATED CLASS',
-          '(deprecated)', ['THIS CLASS IS DEPRECATED. '
-                           'It will be removed in a future version. '])
+          '(deprecated)', [('THIS CLASS IS DEPRECATED. '
+                            'It will be removed in a future version. ')],
+          notice_type='Deprecated')
       __name__ = func_or_class.__name__
       __module__ = _call_location(outer=True)
 
-      @_wrap_decorator(func_or_class.__init__)
+      @_wrap_decorator(func_or_class.__init__, 'deprecated_alias')
       def __init__(self, *args, **kwargs):
         if hasattr(_NewClass.__init__, '__func__'):
           # Python 2
@@ -222,7 +245,7 @@ def deprecated_alias(deprecated_name, name, func_or_class, warn_once=True):
           if _NewClass.__init__ not in _PRINTED_WARNING:
             if warn_once:
               _PRINTED_WARNING[_NewClass.__init__] = True
-            logging.warning(
+            _log_deprecation(
                 'From %s: The name %s is deprecated. Please use %s instead.\n',
                 _call_location(), deprecated_name, name)
         super(_NewClass, self).__init__(*args, **kwargs)
@@ -241,10 +264,11 @@ def deprecated_alias(deprecated_name, name, func_or_class, warn_once=True):
         if new_func not in _PRINTED_WARNING:
           if warn_once:
             _PRINTED_WARNING[new_func] = True
-          logging.warning(
+          _log_deprecation(
               'From %s: The name %s is deprecated. Please use %s instead.\n',
               _call_location(), deprecated_name, name)
       return func_or_class(*args, **kwargs)
+
     return tf_decorator.make_decorator(
         func_or_class, new_func, 'deprecated',
         _add_deprecated_function_notice_to_docstring(
@@ -267,16 +291,18 @@ def deprecated_endpoints(*args):
     _tf_deprecated_api_names would be set to a list of deprecated
     endpoint names for the symbol.
   """
+
   def deprecated_wrapper(func):
     # pylint: disable=protected-access
     if '_tf_deprecated_api_names' in func.__dict__:
-      raise DeprecatedNamesAlreadySet(
+      raise DeprecatedNamesAlreadySetError(
           f'Cannot set deprecated names for {func.__name__} to {args}. '
           'Deprecated names are already set to '
           f'{func._tf_deprecated_api_names}.')
     func._tf_deprecated_api_names = args
     # pylint: disable=protected-access
     return func
+
   return deprecated_wrapper
 
 
@@ -298,8 +324,8 @@ def deprecated(date, instructions, warn_once=True):
   to the rest of the docstring.
 
   Args:
-    date: String or None. The date the function is scheduled to be removed.
-      Must be ISO 8601 (YYYY-MM-DD), or None.
+    date: String or None. The date the function is scheduled to be removed. Must
+      be ISO 8601 (YYYY-MM-DD), or None.
     instructions: String. Instructions on how to update code using the
       deprecated function.
     warn_once: Boolean. Set to `True` to warn only the first time the decorated
@@ -320,8 +346,14 @@ def deprecated(date, instructions, warn_once=True):
       # If a class is deprecated, you actually want to wrap the constructor.
       cls = func_or_class
       if cls.__new__ is object.__new__:
+        # If a class defaults to its parent's constructor, wrap that instead.
         func = cls.__init__
         constructor_name = '__init__'
+        decorators, _ = tf_decorator.unwrap(func)
+        for decorator in decorators:
+          if decorator.decorator_name == 'deprecated':
+            # If the parent is already deprecated, there's nothing to do.
+            return cls
       else:
         func = cls.__new__
         constructor_name = '__new__'
@@ -332,17 +364,20 @@ def deprecated(date, instructions, warn_once=True):
       func = func_or_class
 
     decorator_utils.validate_callable(func, 'deprecated')
-    @functools.wraps(func)
+
+    @_wrap_decorator(func, 'deprecated')
     def new_func(*args, **kwargs):  # pylint: disable=missing-docstring
       if _PRINT_DEPRECATION_WARNINGS:
-        if func not in _PRINTED_WARNING:
+        if func not in _PRINTED_WARNING and cls not in _PRINTED_WARNING:
           if warn_once:
             _PRINTED_WARNING[func] = True
-          logging.warning(
+            if cls:
+              _PRINTED_WARNING[cls] = True
+          _log_deprecation(
               'From %s: %s (from %s) is deprecated and will be removed %s.\n'
-              'Instructions for updating:\n%s',
-              _call_location(), decorator_utils.get_qualified_name(func),
-              func.__module__,
+              'Instructions for updating:\n%s', _call_location(),
+              decorator_utils.get_qualified_name(func),
+              func_or_class.__module__,
               'in a future version' if date is None else ('after %s' % date),
               instructions)
       return func(*args, **kwargs)
@@ -352,6 +387,7 @@ def deprecated(date, instructions, warn_once=True):
         func, new_func, 'deprecated',
         _add_deprecated_function_notice_to_docstring(func.__doc__, date,
                                                      instructions))
+    new_func.__signature__ = inspect.signature(func)
 
     if cls is None:
       return new_func
@@ -391,18 +427,18 @@ def deprecated_args(date, instructions, *deprecated_arg_names_or_tuples,
   prepended to the rest of the docstring.
 
   Args:
-    date: String or None. The date the function is scheduled to be removed.
-      Must be ISO 8601 (YYYY-MM-DD), or None.
+    date: String or None. The date the function is scheduled to be removed. Must
+      be ISO 8601 (YYYY-MM-DD), or None.
     instructions: String. Instructions on how to update code using the
       deprecated function.
-    *deprecated_arg_names_or_tuples: String or 2-Tuple (String,
-      ok_val).  The string is the deprecated argument name.
-      Optionally, an ok-value may be provided.  If the user provided
-      argument equals this value, the warning is suppressed.
+    *deprecated_arg_names_or_tuples: String or 2-Tuple (String, ok_val).  The
+      string is the deprecated argument name. Optionally, an ok-value may be
+      provided.  If the user provided argument equals this value, the warning is
+      suppressed.
     **kwargs: If `warn_once=False` is passed, every call with a deprecated
       argument will log a warning. The default behavior is to only warn the
-      first time the function is called with any given deprecated argument.
-      All other kwargs raise `ValueError`.
+      first time the function is called with any given deprecated argument. All
+      other kwargs raise `ValueError`.
 
   Returns:
     Decorated function or method.
@@ -443,8 +479,8 @@ def deprecated_args(date, instructions, *deprecated_arg_names_or_tuples,
          suppressed.
 
     Args:
-      names_to_ok_vals: dict from string arg_name to a list of values,
-        possibly empty, which should not elicit a warning.
+      names_to_ok_vals: dict from string arg_name to a list of values, possibly
+        empty, which should not elicit a warning.
       arg_spec: Output from tf_inspect.getfullargspec on the called function.
 
     Returns:
@@ -452,8 +488,7 @@ def deprecated_args(date, instructions, *deprecated_arg_names_or_tuples,
     """
     # Extract argument list
     arg_space = arg_spec.args + arg_spec.kwonlyargs
-    arg_name_to_pos = {
-        name: pos for pos, name in enumerate(arg_space)}
+    arg_name_to_pos = {name: pos for pos, name in enumerate(arg_space)}
     deprecated_positional_args = {}
     for arg_name, spec in iter(names_to_ok_vals.items()):
       if arg_name in arg_name_to_pos:
@@ -475,14 +510,15 @@ def deprecated_args(date, instructions, *deprecated_arg_names_or_tuples,
     is_varargs_deprecated = arg_spec.varargs in deprecated_arg_names
     is_kwargs_deprecated = arg_spec.varkw in deprecated_arg_names
 
-    if (len(deprecated_positions) + is_varargs_deprecated
-        + is_kwargs_deprecated
+    if (len(deprecated_positions) + is_varargs_deprecated + is_kwargs_deprecated
         != len(deprecated_arg_names_or_tuples)):
-      known_args = (arg_spec.args
-                    + arg_spec.kwonlyargs
-                    + [arg_spec.varargs, arg_spec.varkw])
-      missing_args = [arg_name for arg_name in deprecated_arg_names
-                      if arg_name not in known_args]
+      known_args = (
+          arg_spec.args + arg_spec.kwonlyargs +
+          [arg_spec.varargs, arg_spec.varkw])
+      missing_args = [
+          arg_name for arg_name in deprecated_arg_names
+          if arg_name not in known_args
+      ]
       raise ValueError('The following deprecated arguments are not present '
                        f'in the function signature: {missing_args}. '
                        'Expected arguments from the following list: '
@@ -542,7 +578,7 @@ def deprecated_args(date, instructions, *deprecated_arg_names_or_tuples,
           if (func, arg_name) not in _PRINTED_WARNING:
             if warn_once:
               _PRINTED_WARNING[(func, arg_name)] = True
-            logging.warning(
+            _log_deprecation(
                 'From %s: calling %s (from %s) with %s is deprecated and will '
                 'be removed %s.\nInstructions for updating:\n%s',
                 _call_location(), decorator_utils.get_qualified_name(func),
@@ -558,7 +594,9 @@ def deprecated_args(date, instructions, *deprecated_arg_names_or_tuples,
   return deprecated_wrapper
 
 
-def deprecated_arg_values(date, instructions, warn_once=True,
+def deprecated_arg_values(date,
+                          instructions,
+                          warn_once=True,
                           **deprecated_kwargs):
   """Decorator for marking specific function argument values as deprecated.
 
@@ -577,8 +615,8 @@ def deprecated_arg_values(date, instructions, warn_once=True,
   prepended to the rest of the docstring.
 
   Args:
-    date: String or None. The date the function is scheduled to be removed.
-      Must be ISO 8601 (YYYY-MM-DD), or None
+    date: String or None. The date the function is scheduled to be removed. Must
+      be ISO 8601 (YYYY-MM-DD), or None
     instructions: String. Instructions on how to update code using the
       deprecated function.
     warn_once: If `True`, warn only the first time this function is called with
@@ -600,6 +638,7 @@ def deprecated_arg_values(date, instructions, warn_once=True,
   def deprecated_wrapper(func):
     """Deprecation decorator."""
     decorator_utils.validate_callable(func, 'deprecated_arg_values')
+
     @functools.wraps(func)
     def new_func(*args, **kwargs):
       """Deprecation wrapper."""
@@ -611,16 +650,18 @@ def deprecated_arg_values(date, instructions, warn_once=True,
             if (func, arg_name) not in _PRINTED_WARNING:
               if warn_once:
                 _PRINTED_WARNING[(func, arg_name)] = True
-              logging.warning(
+              _log_deprecation(
                   'From %s: calling %s (from %s) with %s=%s is deprecated and '
                   'will be removed %s.\nInstructions for updating:\n%s',
                   _call_location(), decorator_utils.get_qualified_name(func),
-                  func.__module__, arg_name, arg_value, 'in a future version'
-                  if date is None else ('after %s' % date), instructions)
+                  func.__module__, arg_name, arg_value,
+                  'in a future version' if date is None else
+                  ('after %s' % date), instructions)
       return func(*args, **kwargs)
 
-    doc = _add_deprecated_arg_value_notice_to_docstring(
-        func.__doc__, date, instructions, deprecated_kwargs)
+    doc = _add_deprecated_arg_value_notice_to_docstring(func.__doc__, date,
+                                                        instructions,
+                                                        deprecated_kwargs)
     return tf_decorator.make_decorator(func, new_func, 'deprecated', doc)
 
   return deprecated_wrapper
@@ -634,6 +675,7 @@ def deprecated_argument_lookup(new_name, new_value, old_name, old_value):
     new_value: value of new argument (or None if not used)
     old_name: old name of argument
     old_value: value of old argument (or None if not used)
+
   Returns:
     The effective argument that should be used.
   Raises:
@@ -647,8 +689,9 @@ def deprecated_argument_lookup(new_name, new_value, old_name, old_value):
 
 
 def rewrite_argument_docstring(old_doc, old_argument, new_argument):
-  return old_doc.replace('`%s`' % old_argument, '`%s`' % new_argument).replace(
-      '%s:' % old_argument, '%s:' % new_argument)
+  return old_doc.replace('`%s`' % old_argument,
+                         '`%s`' % new_argument).replace('%s:' % old_argument,
+                                                        '%s:' % new_argument)
 
 
 @tf_contextlib.contextmanager
@@ -661,6 +704,42 @@ def silence():
   _PRINT_DEPRECATION_WARNINGS = print_deprecation_warnings
 
 
+def deprecate_moved_module(deprecated_name, new_module, deletion_version):
+  """Logs a warning when a module that has been moved to a new location is used.
+
+  Copy the following code into the old module:
+
+  ```
+  import deprecation
+  import new_module
+
+  __getattr__ = deprecation.deprecate_moved_module(
+      __name__, new_module, "2.9")  # adjust version number.
+  ```
+
+  Args:
+    deprecated_name: Name of old module.
+    new_module: Module to replace the old module.
+    deletion_version: Version of TensorFlow in which the old module will be
+      removed.
+
+  Returns:
+    A function that logs a warning and returns the symbol from the new module.
+    Set this function as the module's `__getattr__`.
+  """
+
+  def getter(name):
+    if getter not in _PRINTED_WARNING and _PRINT_DEPRECATION_WARNINGS:
+      _PRINTED_WARNING[getter] = True
+      _log_deprecation(
+          'Please fix your imports. Module %s has been moved to %s. The old '
+          'module will be deleted in version %s.', deprecated_name,
+          new_module.__name__, deletion_version)
+    return getattr(new_module, name)
+
+  return getter
+
+
 class HiddenTfApiAttribute(property):
   """Hides a class attribute from the public API.
 
@@ -668,10 +747,6 @@ class HiddenTfApiAttribute(property):
   front of the name (e.g. ClassName._variables). This doesn't work when
   attributes or methods are inherited from a parent class. To hide inherited
   attributes, set their values to be `deprecation.hide_attribute_from_api`.
-  For example, this is used in V2 Estimator to hide the deprecated
-  export_savedmodel method:
-    class EstimatorV2(Estimator):
-       export_savedmodel = deprecation.hide_attribute_from_api('...')
   """
 
   def __init__(self, deprecation_message):
